package twcsckernel.serverKernel.impl;

import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import twcsckernel.projectbase.common.FsChangeListener;
import twcsckernel.projectbase.common.RemoteAgent;
import twcsckernel.projectbase.common.RemoteFileSystem;
import twcsckernel.projectbase.common.User;
import twcsckernel.projectbase.io.FileSecurityManager;
import twcsckernel.projectbase.io.RemoteFileSystemImpl;
import twcsckernel.projectbase.plugins.PluginDescriptor;
import twcsckernel.serverKernel.io.LogWriter;
import twcsckernel.serverKernel.io.SharedFsChangeListener;
import twcsckernel.serverKernel.plugins.PluginHandle;
import twcsckernel.serverKernel.plugins.ServerPlugin;
import twcsckernel.serverKernel.usr.LoginException;
import twcsckernel.serverKernel.utils.RmiManager;
import twcsckernel.serverKernel.utils.ServerGlobalState;
import twcsckernel.serverKernel.utils.UserData;

/**
 * Implementacja interfejsu uytkownika.
 * 
 * @author VMD Group
 * 
 */
public class UserImpl implements User {

	private UserData userData = null;

	private ServerGlobalState globalState;

	private Map<String, ServerPlugin> loggedPlugins;

	private Map<String, RemoteAgent> registeredAgents;

	private SharedFsChangeListener sharedFsChangeListener;

	private FsChangeListener currentRemoteFsListener = null;

	private long timeStamp;

	private RmiManager rmiManager;

	private RemoteFileSystemImpl remoteFileSystem;

	private FileSecurityManager secManager;

	private String userHost;

	public UserImpl(UserData myself, ServerGlobalState globalState,
			RmiManager rmiMan, String userHost) throws IOException {
		userData = myself;
		rmiManager = rmiMan;
		this.userHost = userHost;
		try {
			this.secManager = globalState.loggedUsers.login(this);
		} catch (LoginException e) {
			// w konstruktorze MULTIPLE_LOGIN nie wystapi;)
			throw new IOException(
					"An error occured while creating security manager");
		}
		remoteFileSystem=new RemoteFileSystemImpl(secManager);

		rmiManager.exportObject(remoteFileSystem.writerFactory);
		try {
			rmiManager.exportObject(remoteFileSystem.readerFactory);
		} catch (RemoteException re) {
			globalState.loggedUsers.logout(this);
			rmiManager.unexportObject(remoteFileSystem.writerFactory, true);
			throw re;
		}
		try {
			rmiManager.exportObject(remoteFileSystem.fileFactory);
		} catch (RemoteException re) {
			globalState.loggedUsers.logout(this);
			rmiManager.unexportObject(remoteFileSystem.writerFactory, true);
			rmiManager.unexportObject(remoteFileSystem.readerFactory, true);
			throw re;
		}
		try {
			rmiManager.exportObject(remoteFileSystem);
		} catch (RemoteException re) {
			globalState.loggedUsers.logout(this);
			rmiManager.unexportObject(remoteFileSystem.writerFactory, true);
			rmiManager.unexportObject(remoteFileSystem.readerFactory, true);
			rmiManager.unexportObject(remoteFileSystem.fileFactory, true);
			throw re;
		}
		sharedFsChangeListener = globalState.loggedUsers
				.getSharedUserListener(this);
		remoteFileSystem.addFilesystemChangeListener(sharedFsChangeListener);
		this.globalState = globalState;
		loggedPlugins = new HashMap<String, ServerPlugin>();
		registeredAgents = new HashMap<String, RemoteAgent>();
		timeStamp = System.currentTimeMillis();
	}

	// Metody zdalne

	public PluginDescriptor[] listClientPlugins() throws RemoteException {
		ping();
		return globalState.pluginHandleContainer
				.getUserPluginsDescriptors(userData.getType());
	}

	public byte[] downloadClientPlugin(String clientPackagePath)
			throws RemoteException, IOException {
		ping();
		PluginHandle pHandle = globalState.pluginHandleContainer
				.getUserPluginByClientPackagePath(userData.getType(),
						clientPackagePath);
		if (pHandle == null)
			return null;
		return pHandle.getClientPluginData();
	}

	public RemoteAgent getRemotePluginAgent(String clientPackagePath,
			RemoteAgent clientCallback) throws RemoteException {
		ping();
		RemoteAgent agent;
		synchronized (registeredAgents) {
			agent = registeredAgents.get(clientPackagePath);
			if (agent == null) {
				PluginHandle plugin = globalState.pluginHandleContainer
						.getUserPluginByClientPackagePath(userData.getType(),
								clientPackagePath);
				if (plugin == null)
					throw new RemoteException("No such plugin: "
							+ clientPackagePath);
				ServerPlugin pluginInstance;
				try {
					pluginInstance = plugin.serverPluginBuilder.getInstance(
							this, clientCallback, globalState);
				} catch (Exception e) {
					throw new RemoteException("exception: "
							+ e.getClass().getName() + " message: "
							+ e.getMessage(), e.getCause());
				}
				agent = pluginInstance.getOwnAgent(this);
				loggedPlugins.put(clientPackagePath, pluginInstance);
				rmiManager.exportObject(agent);
				registeredAgents.put(clientPackagePath, agent);
			}
		}
		return agent;
	}

	public void logout() throws RemoteException {
		try {
			logout0();
		} catch (RemoteException re) {
			globalState.logWriter.writeUserLog(
					"Error while invoking remote logout of user: \""
							+ userData.getLogin() + "\" (" + userHost + ")",
					LogWriter.ERROR_OTHER);
			throw re;
		}
		globalState.logWriter.writeUserLog("User \"" + userData.getLogin()
				+ "\" (" + userHost + ") has successfully logged out.",
				LogWriter.USER_REMOTE_LOGOUT);
	}

	public RemoteFileSystem getUserFileSystem() throws RemoteException {
		return remoteFileSystem;
	}

	public void setFsChangeListener(FsChangeListener fsChangeListener)
			throws RemoteException {
		ping();
		// pewne implementacje mog nie obsugiwa nasuchiwania plikw
		if (sharedFsChangeListener == null)
			return;
		synchronized (sharedFsChangeListener) {
			if (currentRemoteFsListener != null)
				sharedFsChangeListener.removeFsChangeListener(
						currentRemoteFsListener, this);
			if (fsChangeListener != null)
				sharedFsChangeListener.addFsChangeListener(fsChangeListener,
						this, secManager.getUserRootPathManager());
			currentRemoteFsListener = fsChangeListener;
		}
	}

	public void ping() throws RemoteException {
		timeStamp = System.currentTimeMillis();
	}

	public long getActivityTimeout() throws RemoteException {
		return rmiManager.getUserCollectorInterval();
	}

	// Metody lokalne

	private void logout0() throws RemoteException {
		/*
		 * synchronizacje na agentach i pluginach zapewnia unexportObject(this)
		 * odcinajac zdalny dostep do obiektu, a przy powtornym wylogowaniu w
		 * przypadku garbage collecora wyrzucony zostanie wyjatek
		 * NoSuchObjectException, ktrory takze w tym przypadku zapewni
		 * synchronizacje
		 */

		rmiManager.unexportObject(this, true);
		rmiManager.unexportObject(remoteFileSystem, true);
		rmiManager.unexportObject(remoteFileSystem.fileFactory, true);
		rmiManager.unexportObject(remoteFileSystem.readerFactory, true);
		rmiManager.unexportObject(remoteFileSystem.writerFactory, true);
		globalState.loggedUsers.logout(this);
		Collection<ServerPlugin> loggedPluginsCollection = loggedPlugins
				.values();
		for (ServerPlugin srvPlugin : loggedPluginsCollection)
			srvPlugin.logoutUser(this);
		Collection<RemoteAgent> registeredAgentsCollection = registeredAgents
				.values();
		for (RemoteAgent agent : registeredAgentsCollection)
			rmiManager.unexportObject(agent, true);

		remoteFileSystem.releaseAllResources();

	}

	/**
	 * Metoda lokalnego wylogowania uytkownika.
	 */
	public void kickUser() {
		try {
			logout0();
		} catch (RemoteException re) {
			globalState.logWriter.writeUserLog(
					"Error while invoking local logout [kick] of user: \""
							+ userData.getLogin() + "\" (owner: " + userHost
							+ ")", LogWriter.ERROR_OTHER);
		}
		globalState.logWriter.writeUserLog("User \"" + userData.getLogin()
				+ "\" (owner: " + userHost
				+ ") has been successfully kicked out.",
				LogWriter.USER_KICK_LOGOUT);
	}

	/**
	 * @return - login uytkownika
	 */
	public String getUserLogin() {
		return userData.getLogin();
	}

	/**
	 * @return - warto znacznika czasowego ostatniej operacji jakiej dokonal
	 *         uytkownik (<i>System.currentTimeMilis()</i>)
	 */
	public long getTimeStamp() {
		return timeStamp;
	}

	/**
	 * @return - cieka root uytkownika
	 */
	public String getUserRootPath() {
		return userData.getRoot();
	}

	/**
	 * @return - typ uytkownika
	 */
	public String getUserType() {
		return userData.getType();
	}

	/**
	 * @return - pene dane uytkownika
	 */
	public UserData getFullUserData() {
		return userData;
	}

}
