package twcsckernel.serverKernel.utils;

import java.io.IOException;
import java.rmi.NoSuchObjectException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.RMISocketFactory;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import twcsckernel.projectbase.common.RemoteAgent;
import twcsckernel.projectbase.common.RemoteServer;
import twcsckernel.projectbase.rmi.registry.LocateRegistry;
import twcsckernel.projectbase.rmi.registry.RegistryImpl;
import twcsckernel.projectbase.rmi.socketfactory.ServerTwoWaySocketFactory;
import twcsckernel.serverKernel.impl.UserImpl;
import twcsckernel.serverKernel.usr.UserGarbageCollector;

/**
 * Manager serwera RMI sucy do odpalania serwera RMI wraz z przydzieleniem
 * serwerowi i rejestracj podanej implementacji
 * {@link twcsckernel.projectbase.common.RemoteServer RemoteServer}.<br/>
 * Ponadto wycznie ta klasa powinna by uywana do rejestrowania i
 * wyrejestrowywania obiektw zdalnych z serwera RMI za pomoc metod
 * {@link #exportObject(Remote) exportObject} oraz
 * {@link #unexportObject(Remote, boolean) unexportObject}
 * 
 * @author VMD Group
 * 
 */
public class RmiManager {

	private RegistryImpl registry = null;

	private Map<RemoteAgent, Integer> registeredAgentMap = new HashMap<RemoteAgent, Integer>();

	private Set<UserImpl> registeredUserSet = new HashSet<UserImpl>();

	private UserGarbageCollector userGbC = null;

	private int port;

	private long gbcInterval;

	/**
	 * Konstruktor rejestrujcy odpowiedni fabryk gniazd sieciowych.
	 * 
	 * @param port -
	 *            numer portu na ktrym bdzie odpalany serwer RMI
	 * @param useUserGbC -
	 *            <code>true</code> jeli ma by wczony user garbage
	 *            collector lub <code>false</code> w przeciwnym wypadku
	 * @param connectionTimeout -
	 *            timeout (w ms) prby poczenia do klienta (np. jeli jest
	 *            za NAT/firewall). Jeli zero, to traktowany jest jak 
	 *            nieskoczono i poczenie bdzie nawizywane dopki nie
	 *            wystpi bd poczenia (wtedy dopiero natstpuje poczenie
	 *            pasywne).
	 * @throws IOException -
	 *             wyjtek rzucany w przypadku niepowodzenia rejestracji fabryki
	 *             gniazd sieciowych.
	 */
	public RmiManager(int port, boolean useUserGbC, int connectionTimeout)
			throws IOException {
		RMISocketFactory.setSocketFactory(new ServerTwoWaySocketFactory(
				connectionTimeout));
		this.port = port;
		if (useUserGbC) {
			userGbC = new UserGarbageCollector();
			userGbC.setUserSet(registeredUserSet);
		}
	}

	/**
	 * @return interwa czasu (w milisekunach) co jaki ma by "sprztane" - zero
	 *         oznacza wyczony user garbage collector
	 */
	public long getUserCollectorInterval() {
		return gbcInterval;
	}

	/**
	 * Ustawia "sprztacza" uytkownikw, ktrzy nie utrzymywali poczenia
	 * przez pewnien okrelony w nim czas.
	 * 
	 * @param interval -
	 *            interwa czasu (w milisekunach) co jaki ma by "sprztane" -
	 *            zero oznacza wyczenie
	 */
	public void setUserCollectorInterval(long interval) {
		if (userGbC != null) {
			userGbC.setInterval(interval);
			gbcInterval = interval;
		}
	}

	/**
	 * Metoda odpala serwer na danym porcie oraz rejestruje podany serwer
	 * logowania
	 * {@link twcsckernel.projectbase.common.RemoteServer RemoteServer} na
	 * serwerze RMI.
	 * 
	 * @param remoteServer -
	 *            serwer logowania rejestrowany w serwerze RMI
	 * @throws RemoteException -
	 *             wyjtek rzucany w przypadku niepowodzenia utworzenia serwera
	 *             RMI lub niepowodzenia rejestracji na nim serwera logowania
	 */
	public void startRegistryServer(RemoteServer remoteServer)
			throws RemoteException {
		registry = (RegistryImpl) LocateRegistry.createRegistry(port);
		try {
			UnicastRemoteObject.exportObject(remoteServer, port);
		} catch (RemoteException re) {
			UnicastRemoteObject.unexportObject(registry, true);
			registry = null;
			throw re;
		}
		registry.registerRemoteServer(remoteServer);
		if (userGbC != null)
			userGbC.enable();
	}

	/**
	 * Metoda wyczajca serwer RMI.
	 * 
	 * @return <code>true</code> wtedy i tylko wtedy gdy serwer by wczony i
	 *         operacja wyczenia si powioda lub <code>false</code> w
	 *         przeciwnym wypadku.
	 */
	public boolean stopRegistryServer() {
		if (registry == null)
			return false;
		try {
			if (userGbC != null)
				userGbC.disable();
			synchronized (registeredUserSet) {
				// wykonywana jest kopia ze wzgledu na to, ze
				// user.kickUser() modyfikuje posrednio
				// registeredUserSet
				UserImpl[] tmpUsrCopy = registeredUserSet
						.toArray(new UserImpl[0]);
				for (UserImpl user : tmpUsrCopy)
					user.kickUser();
				return UnicastRemoteObject.unexportObject(registry, true);
			}
		} catch (NoSuchObjectException e) {
			return false;
		}
	}

	/**
	 * Metoda rejestrujca obiekt zdalny w serwerze RMI. Tylko ona powinna by
	 * uywana do rejestrowania obiektw zdalnych.<br/> Metoda wyrnia dwie
	 * klasy na ktre reaguje oddzielnie:
	 * <ul>
	 * <li><i>UserImpl</i> - metoda zapisuje dodatkowo obiekt w zbiorze
	 * dostpnym dla garbage collectora uytkownikw, przez co podlegaj jego
	 * ew. kontroli.</li>
	 * <li><i>RemoteAgent</i> - obiekty te mona rejestrowa wielokrotnie,
	 * gdy pluginy mog potencjalnie stosowa tego samego agenta dla wielu
	 * uytkownikw. Prowadzony jest zatem dla kadego zarejestrowanego agenta
	 * licznik liczby rejestrowa. Obiekt niezarejestrowany jest rejestrowany i
	 * dodawany do mapy licznikw. Jeli ponownie ten sam agent jest
	 * rejestrowany zwikszany jest tylko licznik rejestrowa.<br/> Obrazowy
	 * schemat:<br/> <code>
	 * if (agent.count==0) {
	 *   rejestruj(agent);
	 *   agent.count=1;
	 * } else
	 *   agent.count++;
	 * </code></li>
	 * </ul>
	 * 
	 * 
	 * @param object -
	 *            rejestrowany obiekt
	 * @throws RemoteException -
	 *             wyjtek rzucany w przypadku niepowodzenia rejestracji
	 */
	public void exportObject(Remote object) throws RemoteException {
		if (object instanceof RemoteAgent)
			synchronized (registeredAgentMap) {
				Integer existing = registeredAgentMap.get(object);
				if (existing == null) {
					existing = 1;
					UnicastRemoteObject.exportObject(object, port);
					registeredAgentMap.put((RemoteAgent) object, existing);
				} else {
					existing++;
					registeredAgentMap.put((RemoteAgent) object, existing);
				}
			}
		else if (object instanceof UserImpl)
			synchronized (registeredUserSet) {
				UnicastRemoteObject.exportObject(object, port);
				registeredUserSet.add((UserImpl) object);
			}
		else
			UnicastRemoteObject.exportObject(object, port);

	}

	/**
	 * Metoda wyrejestrowujca obiekt zdalny z serwera RMI. Tylko ta metoda
	 * powinna by uywana do wyrejestrowania obiektu.<br/> Metoda wyrnia
	 * dwie klasy na ktre reaguje oddzielnie:
	 * <ul>
	 * <li><i>UserImpl</i> - metoda usuwa dodatkowo obiekt ze zbioru
	 * dostpnego dla garbage collectora uytkownikw</li>
	 * <li><i>RemoteAgent</i> - dziaa odwrotnie do
	 * {@link #exportObject(Remote) exportObject}.<br/> Obrazowy schemat:<br/>
	 * <code>
	 * if (agent.count>1)
	 *   agent.count--;
	 * else {
	 *   wyrejestruj(agent);
	 *   agent.count=0;
	 * }
	 * </code>
	 * </li>
	 * </ul>
	 * 
	 * @param object -
	 *            obiekt to wyrejestrowania
	 * @param ifForce -
	 *            jeli <code>true</code> to obiekt zostanie wyrejestrowany
	 *            nawet w przypadku przeprowadzanych w aktualnie operacji
	 *            zdalnych, w przeciwnym wypadku obiekt bdzie wyrejestrowany
	 *            tylko jeli nie wykonuje adnych operacji zdalnych.
	 * @return <code>true</code> jeli operacja zakoczona powodzeniem lub
	 *         <code>false</code> w przeciwnym wypadku.
	 * @throws NoSuchObjectException -
	 *             wyjtek rzucany jeli podany obiekt nie by zarejestrowany
	 */
	public boolean unexportObject(Remote object, boolean ifForce)
			throws NoSuchObjectException {
		if (object instanceof RemoteAgent)
			synchronized (registeredAgentMap) {
				Integer existing = registeredAgentMap.get(object);
				if (existing == null)
					return false; // jak ktos rejestrowal poza rmiManager to
				// niech
				// sie sam martwi
				if (existing > 1) {
					existing--;
					registeredAgentMap.put((RemoteAgent) object, existing);
					return true;
				} else {
					boolean result = UnicastRemoteObject.unexportObject(object,
							ifForce);
					if (result)
						registeredAgentMap.remove(object);
					return result;
				}
			}
		else if (object instanceof UserImpl)
			synchronized (registeredUserSet) {
				boolean result = UnicastRemoteObject.unexportObject(object,
						ifForce);
				if (result)
					registeredUserSet.remove(object);
				return result;
			}
		else
			return UnicastRemoteObject.unexportObject(object, ifForce);
	}
}
