package twcsckernel.projectbase.io;

import java.io.File;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import twcsckernel.projectbase.common.FileFactory;

public class FileFactoryImpl implements FileFactory {

	private Collection<LocalFsChangeListener> fsListeners = null;

	private Map<Integer, File> fileMap = Collections
			.synchronizedMap(new HashMap<Integer, File>());

	private Map<Integer, FileSecurityManager> securityMap = new HashMap<Integer, FileSecurityManager>();

	private static final SecurityException illegalAccess = new SecurityException();

	private static final ItemNotFoundException fileNotFound = new ItemNotFoundException();

	private Object idPointerMutex = new Object();

	private int idPointer = 0;

	private FileSecurityManager defaultSecManager = null;

	public FileFactoryImpl(FileSecurityManager defaultSecManager,
			Collection<LocalFsChangeListener> listeners)
			throws NullPointerException {
		if (defaultSecManager == null || listeners == null)
			throw new NullPointerException();
		this.defaultSecManager = defaultSecManager;
		fsListeners = listeners;
	}

	public RemoteFile newRemoteFile(String path) throws SecurityException,
			RemoteException, IOException {
		return newRemoteFile(path, defaultSecManager);
	}

	public RemoteFile newRemoteFile(String path, FileSecurityManager secManager)
			throws SecurityException, IOException {
		File localFile = secManager.getUserRootPathManager().getFile(path);
		if (secManager.checkRootPath(localFile) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		RemoteFile rFile = null;

		synchronized (idPointerMutex) {
			fileMap.put(idPointer, localFile);
			securityMap.put(idPointer, secManager);
			rFile = new RemoteFile(idPointer, this);
			idPointer++;
		}
		return rFile;
	}

	public boolean disposeFileHandle(int fileID) throws RemoteException {
		boolean result = fileMap.remove(fileID) != null;
		securityMap.remove(fileID);
		return result;
	}

	public char getSeparatorChar() throws RemoteException {
		return File.separatorChar;
	}

	public String getSeparator() throws RemoteException {
		return File.separator;
	}

	public char getPathSeparatorChar() throws RemoteException {
		return File.pathSeparatorChar;
	}

	public String getPathSeparator() throws RemoteException {
		return File.pathSeparator;
	}

	public String getName(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		return file.getName();
	}

	public String getParent(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		File parentFile = file.getParentFile();
		if (secManager.checkRootPath(parentFile) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		return secManager.getUserRootPathManager().getUserAbsolutePath(
				parentFile.getAbsolutePath());
	}

	public String getPath(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		String result = secManager.getUserRootPathManager()
				.getUserAbsolutePath(file.getAbsolutePath());
		return result;
	}

	public boolean canRead(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkReadAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			return false;
		else
			return file.canRead();
	}

	public boolean canWrite(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkWriteAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			return false;
		else
			return file.canWrite();
	}

	public boolean exists(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkReadAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else
			return file.exists();
	}

	public boolean isDirectory(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkReadAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else
			return file.isDirectory();
	}

	public boolean isFile(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkReadAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else
			return file.isFile();
	}

	public boolean isHidden(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkReadAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else
			return file.isHidden();
	}

	public long lastModified(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkReadAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else
			return file.lastModified();
	}

	public long length(int fileID) throws RemoteException, SecurityException,
			ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkReadAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else
			return file.length();
	}

	public boolean createNewFile(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException, IOException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkWriteAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else {
			boolean result = false;
			try {
				result = file.createNewFile();
				if (result) {
					FsChangeDescriptor fd = new FsChangeDescriptor(
							FsChangeDescriptor.FILE_CREATE);
					fd.setChangeInfo(getFileDescriptor(file));
					fireFsListeners(fd);
				}
			} catch (IOException e) {
				throw e;
			}

			return result;
		}
	}

	public boolean delete(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkWriteAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else {
			boolean result = false;
			result = file.delete();
			if (result) {
				FsChangeDescriptor fd = new FsChangeDescriptor(
						FsChangeDescriptor.FILE_DELETE);
				fd.setSource(file.getAbsolutePath());
				fireFsListeners(fd);
			}
			return result;
		}
	}

	public FileDescriptor[] list(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkReadAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else {
			File[] fileList = file.listFiles();
			FileDescriptor fileDsc[] = new FileDescriptor[fileList.length];
			for (int i = 0; i < fileList.length; i++) {
				fileDsc[i] = getFileDescriptor(fileList[i]);
				fileDsc[i].path = secManager.getUserRootPathManager()
						.getUserAbsolutePath(fileDsc[i].path);
			}
			return fileDsc;
		}
	}

	public boolean mkdir(int fileID) throws RemoteException, SecurityException,
			ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkWriteAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else {
			boolean result = false;
			result = file.mkdir();
			if (result) {
				FsChangeDescriptor fd = new FsChangeDescriptor(
						FsChangeDescriptor.FILE_CREATE);
				fd.setChangeInfo(getFileDescriptor(file));
				fireFsListeners(fd);
			}
			return result;
		}
	}

	public boolean mkdirs(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkWriteAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else {
			boolean result = false;
			result = file.mkdirs();
			if (result) {
				FsChangeDescriptor fd = new FsChangeDescriptor(
						FsChangeDescriptor.FILE_MKDIRS);
				fd.setChangeInfo(getFileDescriptor(file));
				fireFsListeners(fd);
			}
			return result;
		}
	}

	public boolean renameTo(int fileID, String newName) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		File destFile = new File(file.getParent() + "/" + newName);
		if (secManager.checkWriteAccess(file) == FileSecurityManager.ACC_ILLEGAL
				|| secManager.checkWriteAccess(destFile) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else {
			boolean result = false;
			String srcPath = file.getAbsolutePath();
			result = file.renameTo(destFile);
			if (result) {
				FsChangeDescriptor fd = new FsChangeDescriptor(
						FsChangeDescriptor.FILE_RENAME);
				fd.setSource(srcPath);
				fd.setTarget(newName);
				fireFsListeners(fd);
			}
			return result;
		}
	}

	public boolean setLastModified(int fileID, long time)
			throws RemoteException, SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkWriteAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else {
			boolean result = false;
			result = file.setLastModified(time);
			if (result) {
				FsChangeDescriptor fd = new FsChangeDescriptor(
						FsChangeDescriptor.FILE_REFRESH);
				fd.setChangeInfo(getFileDescriptor(file));
				fireFsListeners(fd);
			}
			return result;
		}
	}

	public boolean setReadOnly(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		if (secManager.checkWriteAccess(file) == FileSecurityManager.ACC_ILLEGAL)
			throw illegalAccess;
		else {
			boolean result = false;
			result = file.setReadOnly();
			if (result) {
				FsChangeDescriptor fd = new FsChangeDescriptor(
						FsChangeDescriptor.FILE_REFRESH);
				fd.setChangeInfo(getFileDescriptor(file));
				fireFsListeners(fd);
			}
			return result;
		}
	}

	public FileDescriptor getDescriptor(int fileID) throws RemoteException,
			SecurityException, ItemNotFoundException {
		File file = getFile(fileID);
		FileSecurityManager secManager = securityMap.get(fileID);
		FileDescriptor result = getFileDescriptor(file);
		result.path = secManager.getUserRootPathManager().getUserAbsolutePath(
				result.path);
		if (result.path == null) {
			result.path = "";
			result.name = "/";
		}
		return result;
	}

	public void releaseAllResources() {
		fileMap.clear();
	}

	private File getFile(int fileID) throws ItemNotFoundException {
		File file = fileMap.get(fileID);
		if (file == null)
			throw fileNotFound;
		return file;
	}

	private void fireFsListeners(FsChangeDescriptor desc) {
		for (LocalFsChangeListener fsL : fsListeners)
			fsL.fsChangeNotify(desc);
	}

	public static FileDescriptor getFileDescriptor(File file) {
		FileDescriptor result = new FileDescriptor();
		result.name = file.getName();
		result.path = file.getParent();
		result.isFile = file.isFile();
		result.isHidden = file.isHidden();
		result.lastModified = file.lastModified();
		result.length = file.length();
		return result;
	}

}
