package twcsckernel.projectbase.io;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import twcsckernel.projectbase.common.WriterFactory;

public class WriterFactoryImpl implements WriterFactory {

	private Collection<LocalFsChangeListener> fsListeners = null;

	private Map<Integer, FileAndStream> fileAndStreamMap = Collections
			.synchronizedMap(new HashMap<Integer, FileAndStream>());

	private static final FileNotFoundException fileNotFound = new FileNotFoundException();

	private static final SecurityException illegalAccess = new SecurityException();

	private static final ItemNotFoundException streamNotFound = new ItemNotFoundException();

	private Object idPointerMutex = new Object();

	private int idPointer = 0;

	private FileSecurityManager defaultSecManager = null;

	private static class FileAndStream {
		public final File file;
		public final FileOutputStream fStream;
		public FileAndStream(File file, FileOutputStream stream) {
			this.file = file;
			fStream = stream;
		}
	}
	
	public WriterFactoryImpl(FileSecurityManager defaultSecManager,
			Collection<LocalFsChangeListener> listeners)
			throws NullPointerException {
		if (defaultSecManager == null || listeners == null)
			throw new NullPointerException();
		this.defaultSecManager = defaultSecManager;
		fsListeners = listeners;
	}

	public RemoteFileOutputStream newFileOutputStream(String filePath)
			throws RemoteException, FileNotFoundException, SecurityException {
		return newFileOutputStream(filePath, defaultSecManager);
	}

	public RemoteFileOutputStream newFileOutputStream(String filePath,
			FileSecurityManager secManager) throws RemoteException,
			FileNotFoundException, SecurityException {

		File file;
		try {
			file = secManager.getUserRootPathManager().getFile(filePath);
		} catch (IOException e1) {
			throw new FileNotFoundException();
		}
		boolean fileExists = file.exists();
		switch (defaultSecManager.checkWriteAccess(file)) {
			case FileSecurityManager.ACC_NOT_FOUND :
				throw fileNotFound;
			case FileSecurityManager.ACC_ILLEGAL :
				throw illegalAccess;
		}

		RemoteFileOutputStream rfos = null;
		try {
			FileOutputStream fos = new FileOutputStream(file);
			if (!fileExists) // jesli plik nie istnial a udalo sie go
			// utworzyc
			{
				FsChangeDescriptor fd = new FsChangeDescriptor(
						FsChangeDescriptor.FILE_CREATE);
				fd.setChangeInfo(FileFactoryImpl.getFileDescriptor(file));
				fireFsListeners(fd);
			}
			synchronized (idPointerMutex) {
				fileAndStreamMap.put(idPointer, new FileAndStream(file, fos));
				rfos = new RemoteFileOutputStream(idPointer, this);
				idPointer++;
			}
		} catch (FileNotFoundException e) {
			throw e;
		}
		return rfos;
	}

	public void write(int fileID, int b) throws RemoteException, IOException,
			ItemNotFoundException {
		FileOutputStream os = getStream(fileID);
		synchronized (os) {
			os.write(b);
		}
	}

	public void write(int fileID, byte[] b) throws RemoteException,
			IOException, ItemNotFoundException {
		FileOutputStream os = getStream(fileID);
		synchronized (os) {
			os.write(b);
		}
	}

	public void flush(int fileID) throws RemoteException, IOException,
			ItemNotFoundException {
		FileOutputStream os = getStream(fileID);
		synchronized (os) {
			os.flush();
		}

	}

	public void close(int fileID) throws RemoteException, IOException,
			ItemNotFoundException {
		FileAndStream fAndS = fileAndStreamMap.get(fileID);
		if (fAndS == null)
			throw streamNotFound;
		close(fAndS);
		fileAndStreamMap.remove(fileID);
	}
	public void releaseAllResources() {
		Collection<FileAndStream> streams = fileAndStreamMap.values();
		for (FileAndStream fAndS : streams)
			try {
				close(fAndS);
			} catch (IOException e) {
			}
		fileAndStreamMap.clear();
	}

	private void close(FileAndStream fAndS) throws IOException {
		synchronized (fAndS.fStream) {
			fAndS.fStream.close();
			FsChangeDescriptor fd = new FsChangeDescriptor(
					FsChangeDescriptor.FILE_REFRESH);
			fd.setChangeInfo(FileFactoryImpl.getFileDescriptor(fAndS.file));
			fireFsListeners(fd);
		}
	}

	private FileOutputStream getStream(int fileID) throws ItemNotFoundException {
		FileAndStream fAndS = fileAndStreamMap.get(fileID);
		if (fAndS == null)
			throw streamNotFound;
		return fAndS.fStream;
	}
	
	private void fireFsListeners(FsChangeDescriptor desc) {
		for (LocalFsChangeListener fsL : fsListeners)
			fsL.fsChangeNotify(desc);
	}

}
