package twcsckernel.serverKernel.usr;

import java.io.IOException;

import twcsckernel.projectbase.io.FileSecurityManager;
import twcsckernel.serverKernel.impl.UserImpl;
import twcsckernel.serverKernel.io.SharedFsChangeListener;
import twcsckernel.serverKernel.utils.UserData;

/**
 * Klasa jest kontenerem zawierającym listę aktualnie zalogowanych użytkowników
 * oraz służy zarządzaniu listenerami systemu plików (mogą być współdzielone).
 * Użytkownik przy logowaniu pobiera listener metodą
 * {@link #getSharedUserListener(UserImpl) getSharedUserListener}, który
 * przekazuje swoim fabrykom plików. Dodawanie i usuwanie listenerów zdalnych
 * które klient przekaże następuje właśnie poprzez ten listener.
 * 
 * @author VMD Group
 * 
 */
public interface LoggedUsersManager {

	/**
	 * Metoda logująca danego użytkownika.
	 * 
	 * @param user -
	 *            logowany użytkownik
	 * @return - <code>FileSecurityManager</code>
	 * @throws LoginException -
	 *             wyjątek rzucany jeśli użytkowik próbuje się zalogować
	 *             ponownie lub wystąpił błąd podczas tworzenia managera
	 *             bezpieczeństwa
	 */
	public FileSecurityManager login(UserImpl user) throws LoginException;

	/**
	 * Metoda wylogowania danego użytkownika.
	 * 
	 * @param user -
	 *            użytkownik do wylogowania
	 * @return - <code>true</code> jeśli użytkownik był zalogowany i został
	 *         wylogowany lub <code>false</code> w przeciwnym wypadku
	 */
	public boolean logout(UserImpl user);

	/**
	 * @return - liczba wszystkich zalogowanych użytkowników
	 */
	public int getAllUsersNumber();

	/**
	 * @param login -
	 *            login użytkownika
	 * @return - liczba zaogowanych użytkowników o podanym loginie
	 */
	public int getLoginUsersNumber(String login);
	
	/**
	 * @return - zalogowani użytkownicy o podanym loginie
	 */
	public UserImpl[] getLoggedUserType(String login);

	/**
	 * @return - wszyscy zalogowani użytkownicy
	 */
	public UserImpl[] getAllLoggedUsers();

	/**
	 * Dodanie listenera logowania/wylogowania użytkowników. Parametr
	 * <i>UserImpl owner</i> jest potrzebny identyfikacji listenerów które mają
	 * być automatycznie zwolnione podczas wylogowania ich właściciela.
	 * 
	 * @param loginListener -
	 *            listener logowania/wylogowania
	 * @param owner -
	 *            właściciel listenera
	 */
	public void addLoginListener(LoginListener loginListener, UserImpl owner);

	/**
	 * Metoda pozwala na ręczne wyrejestrowanie listenera logowania
	 * 
	 * @param loginListener -
	 *            listener logowania/wylogowania
	 * @param owner -
	 *            właściciel listenera (optymalizacja)
	 * @return - <code>true</code> jeśli listener był zarejestrowany i udało
	 *         się go wyrejestrować lub <code>false</code> w przeciwnym
	 *         wypadku
	 */
	public boolean removeLoginListener(LoginListener loginListener,
			UserImpl owner);

	/**
	 * Metoda służąca do pobierania listenera użytkownika, który będzie
	 * przekazany dalej do fabryki plików. Dla tej samej instancji użytkownika
	 * metoda powinna zwrócić ten sam listener (w przypaku odwołania do tej
	 * metody przez pluginy).
	 * 
	 * @param user -
	 *            użytkonik dla którego listener ma być przydzielony/przekazany
	 * @return - listener
	 */
	public SharedFsChangeListener getSharedUserListener(UserImpl user);

	/**
	 * Metoda tworzy/pobiera manager bezpieczeństwa dla podanego użytkownika.
	 * 
	 * @param userData -
	 *            deskryptor użytkownika
	 * @return - manager bezpieczeństwa przydzielony temu użytkownikowi
	 * @throws IOException -
	 *             wyjątek rzucany w przypadku błędu utworzenia managera dla
	 *             podanego deskryptora
	 */
	public FileSecurityManager createSecurityManager(UserData userData)
			throws IOException;

}
