package twcsckernel.serverKernel.utils;

import java.io.IOException;
import java.rmi.NoSuchObjectException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.server.RMISocketFactory;
import java.rmi.server.UnicastRemoteObject;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import twcsckernel.projectbase.common.RemoteAgent;
import twcsckernel.projectbase.common.RemoteServer;
import twcsckernel.projectbase.rmi.registry.LocateRegistry;
import twcsckernel.projectbase.rmi.registry.RegistryImpl;
import twcsckernel.projectbase.rmi.socketfactory.ServerTwoWaySocketFactory;
import twcsckernel.serverKernel.impl.UserImpl;
import twcsckernel.serverKernel.usr.UserGarbageCollector;

/**
 * Manager serwera RMI służący do odpalania serwera RMI wraz z przydzieleniem
 * serwerowi i rejestracją podanej implementacji
 * {@link twcsckernel.projectbase.common.RemoteServer RemoteServer}.<br/>
 * Ponadto wyłącznie ta klasa powinna być używana do rejestrowania i
 * wyrejestrowywania obiektów zdalnych z serwera RMI za pomocą metod
 * {@link #exportObject(Remote) exportObject} oraz
 * {@link #unexportObject(Remote, boolean) unexportObject}
 * 
 * @author VMD Group
 * 
 */
public class RmiManager {

	private RegistryImpl registry = null;

	private Map<RemoteAgent, Integer> registeredAgentMap = new HashMap<RemoteAgent, Integer>();

	private Set<UserImpl> registeredUserSet = new HashSet<UserImpl>();

	private UserGarbageCollector userGbC = null;

	private int port;

	private long gbcInterval;

	/**
	 * Konstruktor rejestrujący odpowiednią fabrykę gniazd sieciowych.
	 * 
	 * @param port -
	 *            numer portu na którym będzie odpalany serwer RMI
	 * @param useUserGbC -
	 *            <code>true</code> jeśli ma być włączony user garbage
	 *            collector lub <code>false</code> w przeciwnym wypadku
	 * @param connectionTimeout -
	 *            timeout (w ms) próby połączenia do klienta (np. jeśli jest
	 *            za NAT/firewall). Jeśli zero, to traktowany jest jak 
	 *            nieskończoność i połączenie będzie nawiązywane dopóki nie
	 *            wystąpi błąd połączenia (wtedy dopiero natstępuje połączenie
	 *            pasywne).
	 * @throws IOException -
	 *             wyjątek rzucany w przypadku niepowodzenia rejestracji fabryki
	 *             gniazd sieciowych.
	 */
	public RmiManager(int port, boolean useUserGbC, int connectionTimeout)
			throws IOException {
		RMISocketFactory.setSocketFactory(new ServerTwoWaySocketFactory(
				connectionTimeout));
		this.port = port;
		if (useUserGbC) {
			userGbC = new UserGarbageCollector();
			userGbC.setUserSet(registeredUserSet);
		}
	}

	/**
	 * @return interwał czasu (w milisekunach) co jaki ma być "sprzątane" - zero
	 *         oznacza wyłączony user garbage collector
	 */
	public long getUserCollectorInterval() {
		return gbcInterval;
	}

	/**
	 * Ustawia "sprzątacza" użytkowników, którzy nie utrzymywali połączenia
	 * przez pewnien określony w nim czas.
	 * 
	 * @param interval -
	 *            interwał czasu (w milisekunach) co jaki ma być "sprzątane" -
	 *            zero oznacza wyłączenie
	 */
	public void setUserCollectorInterval(long interval) {
		if (userGbC != null) {
			userGbC.setInterval(interval);
			gbcInterval = interval;
		}
	}

	/**
	 * Metoda odpala serwer na danym porcie oraz rejestruje podany serwer
	 * logowania
	 * {@link twcsckernel.projectbase.common.RemoteServer RemoteServer} na
	 * serwerze RMI.
	 * 
	 * @param remoteServer -
	 *            serwer logowania rejestrowany w serwerze RMI
	 * @throws RemoteException -
	 *             wyjątek rzucany w przypadku niepowodzenia utworzenia serwera
	 *             RMI lub niepowodzenia rejestracji na nim serwera logowania
	 */
	public void startRegistryServer(RemoteServer remoteServer)
			throws RemoteException {
		registry = (RegistryImpl) LocateRegistry.createRegistry(port);
		try {
			UnicastRemoteObject.exportObject(remoteServer, port);
		} catch (RemoteException re) {
			UnicastRemoteObject.unexportObject(registry, true);
			registry = null;
			throw re;
		}
		registry.registerRemoteServer(remoteServer);
		if (userGbC != null)
			userGbC.enable();
	}

	/**
	 * Metoda wyłączająca serwer RMI.
	 * 
	 * @return <code>true</code> wtedy i tylko wtedy gdy serwer był włączony i
	 *         operacja wyłączenia się powiodła lub <code>false</code> w
	 *         przeciwnym wypadku.
	 */
	public boolean stopRegistryServer() {
		if (registry == null)
			return false;
		try {
			if (userGbC != null)
				userGbC.disable();
			synchronized (registeredUserSet) {
				// wykonywana jest kopia ze wzgledu na to, ze
				// user.kickUser() modyfikuje posrednio
				// registeredUserSet
				UserImpl[] tmpUsrCopy = registeredUserSet
						.toArray(new UserImpl[0]);
				for (UserImpl user : tmpUsrCopy)
					user.kickUser();
				return UnicastRemoteObject.unexportObject(registry, true);
			}
		} catch (NoSuchObjectException e) {
			return false;
		}
	}

	/**
	 * Metoda rejestrująca obiekt zdalny w serwerze RMI. Tylko ona powinna być
	 * używana do rejestrowania obiektów zdalnych.<br/> Metoda wyróżnia dwie
	 * klasy na które reaguje oddzielnie:
	 * <ul>
	 * <li><i>UserImpl</i> - metoda zapisuje dodatkowo obiekt w zbiorze
	 * dostępnym dla garbage collectora użytkowników, przez co podlegają jego
	 * ew. kontroli.</li>
	 * <li><i>RemoteAgent</i> - obiekty te można rejestrować wielokrotnie,
	 * gdyż pluginy mogą potencjalnie stosować tego samego agenta dla wielu
	 * użytkowników. Prowadzony jest zatem dla każdego zarejestrowanego agenta
	 * licznik liczby rejestrowań. Obiekt niezarejestrowany jest rejestrowany i
	 * dodawany do mapy liczników. Jeśli ponownie ten sam agent jest
	 * rejestrowany zwiększany jest tylko licznik rejestrowań.<br/> Obrazowy
	 * schemat:<br/> <code>
	 * if (agent.count==0) {
	 *   rejestruj(agent);
	 *   agent.count=1;
	 * } else
	 *   agent.count++;
	 * </code></li>
	 * </ul>
	 * 
	 * 
	 * @param object -
	 *            rejestrowany obiekt
	 * @throws RemoteException -
	 *             wyjątek rzucany w przypadku niepowodzenia rejestracji
	 */
	public void exportObject(Remote object) throws RemoteException {
		if (object instanceof RemoteAgent)
			synchronized (registeredAgentMap) {
				Integer existing = registeredAgentMap.get(object);
				if (existing == null) {
					existing = 1;
					UnicastRemoteObject.exportObject(object, port);
					registeredAgentMap.put((RemoteAgent) object, existing);
				} else {
					existing++;
					registeredAgentMap.put((RemoteAgent) object, existing);
				}
			}
		else if (object instanceof UserImpl)
			synchronized (registeredUserSet) {
				UnicastRemoteObject.exportObject(object, port);
				registeredUserSet.add((UserImpl) object);
			}
		else
			UnicastRemoteObject.exportObject(object, port);

	}

	/**
	 * Metoda wyrejestrowująca obiekt zdalny z serwera RMI. Tylko ta metoda
	 * powinna być używana do wyrejestrowania obiektu.<br/> Metoda wyróżnia
	 * dwie klasy na które reaguje oddzielnie:
	 * <ul>
	 * <li><i>UserImpl</i> - metoda usuwa dodatkowo obiekt ze zbioru
	 * dostępnego dla garbage collectora użytkowników</li>
	 * <li><i>RemoteAgent</i> - działa odwrotnie do
	 * {@link #exportObject(Remote) exportObject}.<br/> Obrazowy schemat:<br/>
	 * <code>
	 * if (agent.count>1)
	 *   agent.count--;
	 * else {
	 *   wyrejestruj(agent);
	 *   agent.count=0;
	 * }
	 * </code>
	 * </li>
	 * </ul>
	 * 
	 * @param object -
	 *            obiekt to wyrejestrowania
	 * @param ifForce -
	 *            jeśli <code>true</code> to obiekt zostanie wyrejestrowany
	 *            nawet w przypadku przeprowadzanych w aktualnie operacji
	 *            zdalnych, w przeciwnym wypadku obiekt będzie wyrejestrowany
	 *            tylko jeśli nie wykonuje żadnych operacji zdalnych.
	 * @return <code>true</code> jeśli operacja zakończona powodzeniem lub
	 *         <code>false</code> w przeciwnym wypadku.
	 * @throws NoSuchObjectException -
	 *             wyjątek rzucany jeśli podany obiekt nie był zarejestrowany
	 */
	public boolean unexportObject(Remote object, boolean ifForce)
			throws NoSuchObjectException {
		if (object instanceof RemoteAgent)
			synchronized (registeredAgentMap) {
				Integer existing = registeredAgentMap.get(object);
				if (existing == null)
					return false; // jak ktos rejestrowal poza rmiManager to
				// niech
				// sie sam martwi
				if (existing > 1) {
					existing--;
					registeredAgentMap.put((RemoteAgent) object, existing);
					return true;
				} else {
					boolean result = UnicastRemoteObject.unexportObject(object,
							ifForce);
					if (result)
						registeredAgentMap.remove(object);
					return result;
				}
			}
		else if (object instanceof UserImpl)
			synchronized (registeredUserSet) {
				boolean result = UnicastRemoteObject.unexportObject(object,
						ifForce);
				if (result)
					registeredUserSet.remove(object);
				return result;
			}
		else
			return UnicastRemoteObject.unexportObject(object, ifForce);
	}
}
