package twcsckernel.projectbase.io;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.rmi.RemoteException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import twcsckernel.projectbase.common.ReaderFactory;

public class ReaderFactoryImpl implements ReaderFactory {
	private Map<Integer, FileInputStream> streamMap = Collections
			.synchronizedMap(new HashMap<Integer, FileInputStream>());

	private static final FileNotFoundException fileNotFound = new FileNotFoundException();

	private static final SecurityException illegalAccess = new SecurityException();

	private static final ItemNotFoundException streamNotFound = new ItemNotFoundException();

	private Object idPointerMutex = new Object();

	private int idPointer = 0;

	private FileSecurityManager defaultSecManager = null;

	public ReaderFactoryImpl(FileSecurityManager defaultSecManager)
			throws NullPointerException {
		if (defaultSecManager == null)
			throw new NullPointerException("FileSecurityManager==null");

		this.defaultSecManager = defaultSecManager;
	}

	public RemoteFileInputStream newFileInputStream(String filePath)
			throws RemoteException, FileNotFoundException, SecurityException {
		return newFileInputStream(filePath, defaultSecManager);
	}

	public RemoteFileInputStream newFileInputStream(String filePath,
			FileSecurityManager secManager) throws RemoteException,
			FileNotFoundException, SecurityException {

		File file;
		try {
			file = secManager.getUserRootPathManager().getFile(filePath);
		} catch (IOException e1) {
			throw new FileNotFoundException();
		}
		switch (secManager.checkReadAccess(file)) {
		case FileSecurityManager.ACC_NOT_FOUND:
			throw fileNotFound;
		case FileSecurityManager.ACC_ILLEGAL:
			throw illegalAccess;
		}

		RemoteFileInputStream rfis = null;
		try {
			FileInputStream fis = new FileInputStream(file);
			synchronized (idPointerMutex) {
				streamMap.put(idPointer, fis);
				rfis = new RemoteFileInputStream(idPointer, this);
				idPointer++;
			}
		} catch (FileNotFoundException e) {
			throw e;
		}
		return rfis;
	}

	public ReadResult read(int streamID) throws RemoteException, IOException,
			ItemNotFoundException {
		FileInputStream is = getStream(streamID);
		Integer intResult = null;
		synchronized (is) {
			intResult = is.read();
		}
		return new ReadResult(intResult, null);
	}

	public ReadResult read(int streamID, int len) throws RemoteException,
			IOException, ItemNotFoundException {
		FileInputStream is = getStream(streamID);
		byte[] buffer = new byte[len];
		Integer intResult = null;
		synchronized (is) {
			intResult = is.read(buffer, 0, len);
		}
		return new ReadResult(intResult, buffer);
	}

	public long skip(int streamID, long off) throws RemoteException,
			IOException, ItemNotFoundException {
		FileInputStream is = getStream(streamID);
		long skipResult = 0;
		synchronized (is) {
			skipResult = is.skip(off);
		}
		return skipResult;
	}

	public int available(int streamID) throws RemoteException, IOException,
			ItemNotFoundException {
		FileInputStream is = getStream(streamID);
		int avResult = 0;
		synchronized (is) {
			avResult = is.available();
		}
		return avResult;
	}

	public void close(int streamID) throws RemoteException, IOException,
			ItemNotFoundException {
		FileInputStream is = getStream(streamID);
		synchronized (is) {
			is.close();
			streamMap.remove(streamID);
		}

	}

	public void releaseAllResources() {
		Collection<FileInputStream> streams = streamMap.values();
		for (FileInputStream fStream : streams)
			try {
				synchronized (fStream) {
					fStream.close();
				}
			} catch (IOException e) {
			}
		streamMap.clear();
	}

	private FileInputStream getStream(int streamID)
			throws ItemNotFoundException {
		FileInputStream is = streamMap.get(streamID);
		if (is == null)
			throw streamNotFound;
		return is;
	}

}
