﻿package twcsckernel.projectbase.io;

import java.io.File;
import java.io.IOException;
import java.util.regex.Pattern;

/**
 * Menedżer ścieżki root użytkownika.
 * 
 * @author VMD Group
 * 
 */
public class UserRootPathManager {

	private String rootPath;

	private int rootPathLength;

	// wyr. reg. do zmieniania \ na /
	private static Pattern regex = Pattern.compile("\\\\");

	private static final String slash = "/";

	/**
	 * Konstruktor menedżera ścieżki root. Podana ścieżka musi ustnieć. Ścieżka
	 * jest przekształcana na postać kononiczną (absolutną i unikalną) i w
	 * takiej postaci będzie używana.
	 * 
	 * @param rootPath -
	 *            ścieżka root
	 * @throws IOException -
	 *             wyjątek rzucany w przypadku niepoprawnej ścieżki root
	 */
	public void setUserRootPath(String rootPath) throws IOException {
		File rootFile = new File(rootPath);
		if (!rootFile.exists())
			throw new IOException("Incorrect root path: " + rootPath);
		this.rootPath = rootFile.getCanonicalPath();
		rootPathLength = this.rootPath.length();
	}

	/**
	 * Metoda konwertuje ścieżkę bezwzględną systemu plików na ścieżkę
	 * bezwzględną użytkownika. Separatorem w zwracanej ścieżce zawsze
	 * jest "/".
	 * 
	 * @param globalPath -
	 *            ścieżka bezwzględna systemu plików
	 * @return - ścieżka bezwzględna użytkownika
	 */
	public String getUserAbsolutePath(String globalPath) {
		if (!globalPath.startsWith(rootPath))
			return null;
		String result = correctSlash(globalPath.substring(rootPathLength));
		if (!result.startsWith(slash))
			result = slash + result;
		return result;
	}

	private static String correctSlash(String root) {
		return regex.matcher(root).replaceAll(slash);
	}

	/**
	 * Metoda konwertuje ścieżkę bezwzględną systemu plików na ścieżkę
	 * bezwzględną użytkownika.
	 * 
	 * @param file -
	 *            plik
	 * @return - ścieżka bezwzględna użytkownika
	 */
	public String getUserAbsolutePath(File file) {
		return getUserAbsolutePath(file.getAbsolutePath());
	}

	/**
	 * Metoda konwertuje ścieżkę bezwzględną użytkownika na ścieżkę bezwzględną
	 * systemu plików.
	 * 
	 * @param userPath -
	 *            ścieżka bezwzględna użytkownika
	 * @return - ścieżka bezwzględna systemu plików
	 */
	public String getGlobalAbsolutePath(String userPath) {
		return correctSlash(rootPath + slash + userPath);
	}

	/**
	 * Metoda tworzy obiekt pliku mając do dyspozycji bezwzględną ścieżkę
	 * użytkownika.
	 * 
	 * @param userPath -
	 *            bzw. ścieżka użytkownika
	 * @return - obiekt pliku (java.io.File)
	 * @throws IOException -
	 *             wyjątek rzucany jeśli niemożliwe jest uzyskanie kanonicznej
	 *             postaci pliku
	 */
	public File getFile(String userPath) throws IOException {

		return new File(getGlobalAbsolutePath(userPath)).getCanonicalFile();
	}

	/**
	 * @return zwraca ścieżkę główną użytkownika
	 */
	public String getRootPath() {
		return rootPath;
	}
}
