/*
 * Copyright 2000 Computer System Services, Inc.
 *
 * Permission to use this software for any purpose is granted provided that
 * this copyright notice is preserved.
 *
 * This software is provided as-is and without warranty as to its
 * fitness for any purpose.  In other words, Computer System Services,
 * Inc. does not guarantee that this software works.  It is provided
 * only in the hope that it may be found useful by someone.
 *
 * Please e-mail tttaylor@cssassociates.com if you find any errors
 * or want to request changes/enhancements.
 */
package twcsckernel.projectbase.rmi.socketfactory;

import java.rmi.server.*;
import java.net.*;
import java.io.*;
import java.util.*;

/**
 * Server side socket factory which allows the server to make callbacks to a
 * client whose address is unreachable from the server (as long as the server is
 * reachable from the client).
 * 
 * @author Tim Taylor
 */
public class ServerTwoWaySocketFactory extends RMISocketFactory {
	/** Socket pools where sockets requested from client are returned. */
	private Map<String, SocketPool> socketPools = Collections
			.synchronizedMap(new HashMap<String, SocketPool>());

	/** Map of request streams to clients. */
	private Map<String, DataOutputStream> requestStreams = Collections
			.synchronizedMap(new HashMap<String, DataOutputStream>());

	private int timeout;

	/**
	 * Dla inicjacji połączenia do klienta będzie stosowany wskazany timeout.
	 * Jeśli połączenie zostanie ustawione poprawnie timeout jest ustawiany na
	 * zero.
	 * 
	 * @param timeout -
	 *            timeout łączenia z klientem
	 */
	public ServerTwoWaySocketFactory(int timeout) {
		if (timeout < 0)
			timeout = 0;
		this.timeout = timeout;
	}

	/**
	 * Add a request <code>socket</code> to the factory. The factory will use
	 * this <code>socket</code> to request a callback socket from the host of
	 * the given <code>address</code>.
	 * <p>
	 * 
	 * @param address
	 *            Address of the client
	 * @param socket
	 *            Socket on which to request callback sockets from the client.
	 */
	public void addRequestSocket(final byte[] address, Socket socket)
			throws IOException {

		final DataOutputStream requestOutStream = new DataOutputStream(socket
				.getOutputStream());
		final DataInputStream requestInStream = new DataInputStream(socket
				.getInputStream());

		// Thread for receiving notifications of client exports.
		// We need the port to know which request stream to use
		// in the event of multiple client VMs calling from
		// a single host.
		new Thread() {
			public void run() {
				try {
					for (;;) {
						// Read port from data input stream
						int port = requestInStream.readInt();
						String endpoint = EndpointInfo.getEndpointString(
								address, port);

						System.out.println("&&& " + endpoint);

						requestStreams.put(endpoint, requestOutStream);
					}
				} catch (IOException e) {
				}
			}
		}.start();
	}

	/**
	 * Create a socket connectoin to the host of the given <code>address
	 * </code>
	 * and <code>port</code>. If the <code>address</code> is for a host
	 * that has provided a callback socket request signalling channel, then that
	 * channel is used to request that the client create the socket. Otherwise a
	 * socket is created directly to the client from the server.
	 * 
	 * @param address
	 *            Address for the socket.
	 * @param port
	 *            Port for the socket.
	 */
	public Socket createSocket(String address, int port) throws IOException {
		String endpoint = EndpointInfo.getEndpointString(address, port);

		DataOutputStream requestStream = (DataOutputStream) requestStreams
				.get(endpoint);

		if (requestStream == null) {
			try {
				// System.out.println("Jednak tworze socket");
				Socket sock = new Socket();
				sock.connect(new InetSocketAddress(address, port), timeout);
				// return new Socket(address, port);
				if (timeout > 0)
					sock.setSoTimeout(0);
				return sock;
			} catch (IOException ioe) {
			}
		}

		SocketPool requestPool;

		synchronized (socketPools) {
			requestPool = (SocketPool) socketPools.get(endpoint);

			if (requestPool == null) {
				requestPool = new SocketPool();
				socketPools.put(endpoint, requestPool);
			}
		}

		requestStream.writeInt(TwoWay.PROTOCOL_MAGIC);
		requestStream.writeInt(TwoWay.REQUEST_CALLBACK_SOCKET);
		requestStream.writeInt(port);
		requestStream.flush();

		return requestPool.getSocket();
	}

	/**
	 * Private socket class with a buffered input stream. This allows us to use
	 * <code>mark</code> and <code>reset</code> in order to run signalling,
	 * RMI client sockets, and callback sockets through the same sets of
	 * listening ports.
	 */
	private static class BufferedSocket extends Socket {
		Socket sock;
		InputStream in;

		BufferedSocket() throws IOException {
			super();
		}

		public InputStream getInputStream() throws IOException {
			if (in == null) {
				in = new BufferedInputStream(super.getInputStream());
			}

			return in;
		}

		public synchronized void close() throws IOException {
			super.close();
		}
	}

	/**
	 * Server socket that handles both signalling and RMI calls. If a call comes
	 * in with protocol magic in its first bytes, that call is taken to be a
	 * protocol signal and is handled appropriately.
	 * <p>
	 * 
	 * Otherwise, the call is assumed to be an RMI call, and it is passed
	 * through to the "accepting" RMI service.
	 */
	class TwoWayServerSocket extends ServerSocket {
		public TwoWayServerSocket(int port) throws IOException {
			super(port);
		}

		public Socket accept() throws IOException {
			Socket sock = null;

			for (;;) {
				sock = new BufferedSocket();
				implAccept(sock);

				InputStream in = sock.getInputStream();
				in.mark(TwoWay.MAX_MESSAGE_LENGTH);

				DataInputStream dis = new DataInputStream(in);
				DataOutputStream dos = new DataOutputStream(sock
						.getOutputStream());

				int magic = dis.readInt();
				if (magic != TwoWay.PROTOCOL_MAGIC) {
					in.reset();
					break;
				}

				int opcode = dis.readInt();

				if (opcode == TwoWay.REGISTER_CALLBACK_SOCKET_SOURCE) {
					byte[] address = new byte[4];
					dis.read(address, 0, 4);

					addRequestSocket(address, sock);

					// Tell the client our real address and port
					// which may be different from the address
					// and port used to contact us if we are
					// on the other side of a firewall.
					dos.writeInt(TwoWay.RETURN_SERVER_ENDPOINT_INFO);
					dos.write(InetAddress.getLocalHost().getAddress());
					dos.writeInt(sock.getLocalPort());
					dos.flush();
					continue;
				} else if (opcode == TwoWay.RETURN_CALLBACK_SOCKET) {
					byte[] address = new byte[4];
					dis.read(address, 0, 4);
					int port = dis.readInt();

					SocketPool pool = (SocketPool) socketPools.get(EndpointInfo
							.getEndpointString(address, port));
					pool.addSocket(sock);

					continue;
				}

			}

			return sock;
		}
	}
	/**
	 * Returns a server socket for handling the two-way protocol.
	 * 
	 * @param port
	 *            Port on which the socket will listen.
	 */
	public ServerSocket createServerSocket(int port) throws IOException {
		return new TwoWayServerSocket(port);
	}
}
