package put.cs.vmd.server.plugins;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;

import twcsckernel.projectbase.plugins.PluginDescriptor;
import twcsckernel.projectbase.utils.ClassPathLoader;
import twcsckernel.serverKernel.io.LogWriter;
import twcsckernel.serverKernel.plugins.PluginBuilder;
import twcsckernel.serverKernel.plugins.PluginHandle;
import twcsckernel.serverKernel.plugins.PluginHandleContainer;
import twcsckernel.serverKernel.utils.RegisteredTypes;
import twcsckernel.serverKernel.utils.TypeData;

public class PluginManager {

	private String pluginDirPath;

	private RegisteredTypes rTypes;

	private PluginHandleContainer pluginContainer;

	private static final String clientJarName = "client.jar";

	private static final String serverJarName = "server.jar";

	private static final String pluginDescriptorName = "descriptor";

	private LogWriter logWriter;

	public PluginManager(String pluginDirectoryPath, RegisteredTypes rTypes,
			LogWriter logWriter) throws IOException {
		pluginDirPath = pluginDirectoryPath;
		this.rTypes = rTypes;
		pluginContainer = new PluginHandleContainer();
		this.logWriter = logWriter;
		loadPlugins();
	}

	public PluginHandleContainer getPluginHandleContainer() {
		return pluginContainer;
	}

	// zlozonosc suma.po.typach(pluginDirNames.length)*liczba.pluginow
	private void loadPlugins() throws IOException {
		File pluginsDirFile = new File(pluginDirPath);
		if (!pluginsDirFile.isDirectory())
			throw new IOException("Incorrect plugins' directory");
		File[] pluginList = pluginsDirFile.listFiles();
		PluginHandle pluginHandle;
		TypeData[] userTypes = rTypes.getRegisteredTypes();
		for (File pluginDir : pluginList) { // katalog konkretnego plugina
			if (!pluginDir.isDirectory())
				continue;
			pluginHandle = loadPlugin(pluginDir.getPath());
			if (pluginHandle == null)
				continue;
			for (TypeData tData : userTypes) {
				String[] allowedPlugins = tData.getPluginDirNames();
				for (String allowedPlugin : allowedPlugins) {
					if (allowedPlugin.equals(pluginDir.getName()))
						pluginContainer.addPluginToUserType(
								tData.getTypeName(), pluginHandle);
				}
			}
		}

	}

	private PluginHandle loadPlugin(String pluginDir) {
		File serverJar = new File(pluginDir + "/" + serverJarName);
		File clientJar = new File(pluginDir + "/" + clientJarName);
		File descriptorFile = new File(pluginDir + "/" + pluginDescriptorName);

		PluginDescriptor descr = loadDescriptor(descriptorFile);
		PluginBuilder srvPluginBuilder = null;
		if (descr == null) {
			logWriter.writeErrorLog("Could not find descriptor for plugin: "
					+ pluginDir, LogWriter.ERROR_PLUGIN_LOADING);
			return null;
		}

		if (!clientJar.exists()) {
			logWriter.writeErrorLog("Client JAR not found for plugin: "
					+ pluginDir, LogWriter.ERROR_PLUGIN_LOADING);
			return null;
		}

		if (descr.isRemote) {
			try {
				ClassPathLoader.addFile(serverJar);
			} catch (IOException e) {
				logWriter.writeErrorLog(
						"An exception occured while loading plugin classes to JVM: \n"
								+ e.getMessage(),
						LogWriter.ERROR_PLUGIN_LOADING);
				return null;
			}
			Class builderClass = null;
			try {
				builderClass = Class.forName(descr.serverBuilderPackagePath);
			} catch (ClassNotFoundException e) {
				logWriter.writeErrorLog("Required class: '"
						+ descr.serverBuilderPackagePath
						+ "' hasn't been loaded properly: \n" + e.getMessage(),
						LogWriter.ERROR_PLUGIN_LOADING);
				return null;
			}
			try {
				srvPluginBuilder = (PluginBuilder) builderClass.newInstance();
			} catch (Exception e) {
				logWriter.writeErrorLog(
						"An exception occured while instantiating plugin builder: \n"
								+ e.getMessage(),
						LogWriter.ERROR_UNEXPECTED_EXCEPTION);
				return null;
			}
		}
		byte[] clientPluginData = new byte[(int) clientJar.length()];
		FileInputStream fis;
		try {
			fis = new FileInputStream(clientJar);
		} catch (FileNotFoundException fne) {
			return null;
		}
		try {
			fis.read(clientPluginData);
		} catch (IOException ioe) {
			try {
				fis.close();
			} catch (IOException e) {
			}
			return null;
		}
		try {
			fis.close();
		} catch (IOException e) {
		}
		PluginHandle pluginHandle = new PluginHandle(srvPluginBuilder, descr,
				pluginDir, serverJarName, clientJarName, clientPluginData);
		return pluginHandle;
	}

	private PluginDescriptor loadDescriptor(File descriptorFile) {
		InputStream is = openStream(descriptorFile);
		if (is == null)
			return null;
		ObjectInputStream ois = null;
		try {
			ois = new ObjectInputStream(is);
		} catch (IOException e) {
			closeStream(ois);
			return null;
		}
		PluginDescriptor descriptor = null;
		try {
			descriptor = (PluginDescriptor) ois.readObject();
		} catch (Exception e) {
		}
		closeStream(ois);
		return descriptor;
	}

	private InputStream openStream(File file) {
		try {
			return new FileInputStream(file);
		} catch (FileNotFoundException e) {
			return null;
		}
	}

	private void closeStream(InputStream is) {
		try {
			is.close();
		} catch (Exception e) {
		}
	}

}
