package put.cs.vmd.server.usr;

import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import put.cs.vmd.server.io.RootPathSecurityManager;
import put.cs.vmd.server.io.VmdSharedFsChangeListener;
import twcsckernel.projectbase.io.FileSecurityManager;
import twcsckernel.serverKernel.impl.UserImpl;
import twcsckernel.serverKernel.io.SharedFsChangeListener;
import twcsckernel.serverKernel.usr.LoggedUsersManager;
import twcsckernel.serverKernel.usr.LoginException;
import twcsckernel.serverKernel.usr.LoginListener;
import twcsckernel.serverKernel.utils.UserData;

public class VmdLoggedUsersManager implements LoggedUsersManager {

	public static final int USE_USER_SHARED_LISTENER = 0;
	public static final int USE_GLOBAL_SHARED_LISTENER = 1;
	/*
	 * VMD moze obsłużyć jednego listenera współdzielone między użytkownikami
	 * tego samego typu (więc login w mapie wystarcza), lub jednego globalnego
	 * listenera reprezentowanego przez globalSharedListener
	 */
	private Map<String, VmdSharedFsChangeListener> sharedListeners;

	private int sharedFsListenerPolicy;

	private VmdSharedFsChangeListener globalSharedListener;

	// Podwójne definiowanie ze względu na wydajność różnego rodzaju pobierania
	private Map<String, Set<UserImpl>> loggedUsers = new HashMap<String, Set<UserImpl>>();

	private Set<UserImpl> allLoggedUsers = new HashSet<UserImpl>();

	/*
	 * listenery logowania ułożone kategoriami wg użytkowników
	 */
	private Map<UserImpl, Set<LoginListener>> userLoginListenerMap = new HashMap<UserImpl, Set<LoginListener>>();

	/*
	 * listenery logowania zebrane w jeden zbiór - dwukrotne grupowanie ze
	 * wględu na wydajność odpalania
	 */
	private Set<LoginListener> allLoginListenerSet = new HashSet<LoginListener>();

	/*
	 * Operacje na użytkownikach są synchronizowane na loggedUsers, a na
	 * listenerach logowania na userLoginListenerMap
	 */
	public FileSecurityManager login(UserImpl user) throws LoginException {
		String login = user.getUserLogin();
		synchronized (loggedUsers) {
			Set<UserImpl> userSet = loggedUsers.get(login);

			UserData userData = user.getFullUserData();
			FileSecurityManager rootManager;
			try {
				rootManager = createSecurityManager(userData);
			} catch (IOException e) {
				throw new LoginException(LoginException.SECURITY_MANAGER_ERROR);
			}
			if (userSet == null) {
				userSet = new HashSet<UserImpl>();
				loggedUsers.put(login, userSet);
				if (sharedFsListenerPolicy == USE_USER_SHARED_LISTENER)
					sharedListeners.put(login, new VmdSharedFsChangeListener(
							rootManager.getUserRootPathManager()));
			}
			if (userSet.add(user)) {
				allLoggedUsers.add(user);
				fireLoginListers(user, LoginListener.LOGGED_IN);
				return rootManager;
			} else
				throw new LoginException(LoginException.MULTIPLE_LOGIN);
		}
	}

	public boolean logout(UserImpl user) {
		synchronized (loggedUsers) {
			Set<UserImpl> userSet = loggedUsers.get(user.getUserLogin());
			boolean result = userSet.remove(user);
			allLoggedUsers.remove(user);
			if (!result)
				return false;
			if (sharedFsListenerPolicy == USE_GLOBAL_SHARED_LISTENER)
				globalSharedListener.removeUserListeners(user);
			else {
				VmdSharedFsChangeListener userListener = sharedListeners
						.get(user.getUserLogin());
				userListener.removeUserListeners(user);
			}
			removeLoginListeners(user);
			fireLoginListers(user, LoginListener.LOGGED_OUT);
			return true;
		}
	}

	public SharedFsChangeListener getSharedUserListener(UserImpl user) {
		if (sharedFsListenerPolicy == USE_GLOBAL_SHARED_LISTENER)
			return globalSharedListener;
		else
			return sharedListeners.get(user.getUserLogin());
	}

	public UserImpl[] getLoggedUserType(String login) {
		synchronized (loggedUsers) {
			Set<UserImpl> loggedUserType = loggedUsers.get(login);
			if (loggedUserType == null)
				return new UserImpl[0];
			else
				return loggedUserType.toArray(new UserImpl[0]);
		}
	}

	public UserImpl[] getAllLoggedUsers() {
		synchronized (loggedUsers) {
			return allLoggedUsers.toArray(new UserImpl[0]);
		}
	}

	public void addLoginListener(LoginListener loginListener, UserImpl owner) {
		synchronized (userLoginListenerMap) {
			Set<LoginListener> userListenerSet = userLoginListenerMap
					.get(owner);
			if (userListenerSet == null) {
				userListenerSet = new HashSet<LoginListener>();
				userLoginListenerMap.put(owner, userListenerSet);
			}
			userListenerSet.add(loginListener);
			allLoginListenerSet.add(loginListener);
		}
	}

	public boolean removeLoginListener(LoginListener loginListener,
			UserImpl owner) {
		synchronized (userLoginListenerMap) {
			Set<LoginListener> userListenerSet = userLoginListenerMap
					.get(owner);
			if (userListenerSet == null)
				return false;
			userListenerSet.remove(loginListener);
			return allLoginListenerSet.remove(loginListener);
		}
	}

	private void fireLoginListers(UserImpl user, boolean eventType) {
		synchronized (userLoginListenerMap) {
			for (LoginListener lListener : allLoginListenerSet)
				lListener.loginChangeAction(user, eventType);
		}
	}

	private void removeLoginListeners(UserImpl user) {
		synchronized (userLoginListenerMap) {
			Set<LoginListener> userListeners = userLoginListenerMap.get(user);
			if (userListeners == null)
				return;
			allLoginListenerSet.removeAll(userListeners);
			userLoginListenerMap.remove(user);
		}
	}

	public VmdLoggedUsersManager(int sharedFsListenerPolicy) {
		this.sharedFsListenerPolicy = sharedFsListenerPolicy;
		if (sharedFsListenerPolicy == USE_GLOBAL_SHARED_LISTENER)
			globalSharedListener = new VmdSharedFsChangeListener(null);
		else
			sharedListeners = java.util.Collections
					.synchronizedMap(new HashMap<String, VmdSharedFsChangeListener>());
	}

	public FileSecurityManager createSecurityManager(UserData userData)
			throws IOException {
		return new RootPathSecurityManager(userData);
	}

	public int getAllUsersNumber() {
		synchronized (loggedUsers) {
			return allLoggedUsers.size();
		}
	}

	public int getLoginUsersNumber(String login) {
		synchronized (loggedUsers) {
			Set<UserImpl> userSet = loggedUsers.get(login);
			if (userSet == null)
				return 0;
			else
				return userSet.size();
		}
	}
}
